@@ -28,11 +28,20 @@ jm_fit <- function (model_data, model_info, initial_values, priors, control) {
     id_H_ <- rep(idT, each = control$GK_k)
     id_H_ <- match(id_H_, unique(id_H_))
     id_h <- unclass(idT)
+    if (model_data$intgr) {
+        ii <- paste(id_h, model_data$strata, sep = "_")
+        id_h2 <- match(ii, unique(ii))
+        id_h_ <- unlist(mapply2(function (x, ind) if (sum(ind) > 0) x[!duplicated(ind)] else x,
+                                split(id_h, id_h), split(model_data$intgr_ind, id_h)))
+    } else {
+        id_h2 <- id_h_ <- 0
+    }
     model_data <-
         c(model_data, create_Wlong_mats(model_data, model_info,
                                         initial_values, priors,
                                         control),
-          list(id_H = id_H, id_H_ = id_H_, id_h = id_h))
+          list(id_H = id_H, id_H_ = id_H_, id_h = id_h, id_h2 = id_h2,
+               id_h_ = id_h_))
     # cbind the elements of X_H and Z_H, etc.
     model_data$X_H[] <- lapply(model_data$X_H, docall_cbind)
     model_data$X_h[] <- lapply(model_data$X_h, docall_cbind)
@@ -138,7 +147,10 @@ jm_fit <- function (model_data, model_info, initial_values, priors, control) {
     }
     # reconstruct D matrix
     get_D <- function (x) {
-        mapply2(reconstr_D, split(x$L, row(x$L)), split(x$sds, row(x$sds)))
+        ll <- split(x$L, row(x$L))
+        if (!length(ll)) ll <- vector("list", nrow(x$L))
+        mapply2(reconstr_D, ll, split(x$sds, row(x$sds)),
+                MoreArgs = list(ind_zero_D = model_data$ind_zero_D))
     }
     for (i in seq_along(out)) {
         out[[i]][["mcmc"]][["D"]] <-
@@ -169,9 +181,9 @@ jm_fit <- function (model_data, model_info, initial_values, priors, control) {
                    seq_along(out[[i]][["mcmc"]][["sigmas"]][1, ]))
         colnames(out[[i]][["mcmc"]][["sigmaF"]]) <- "sigma_frailty"
         colnames(out[[i]][["mcmc"]][["alphaF"]]) <-
-          paste0("frailty:", 
-                 names(model_info$frames$mf_Surv)[attr(model_info$terms$terms_Surv, 
-                                                       "specials")$strata], 
+          paste0("frailty:",
+                 names(model_info$frames$mf_Surv)[attr(model_info$terms$terms_Surv,
+                                                       "specials")$strata],
                  levels(model_data$strata)[-1]
         )
     }
@@ -330,7 +342,7 @@ jm_fit <- function (model_data, model_info, initial_values, priors, control) {
         c(mlogLik_jm(thetas, statistics$Mean[["b"]], statistics$post_vars,
                      model_data, model_info, control))
     marginal_fit_stats <- fit_stats(mcmc_out$mlogLik, mlogLik_mean_parms)
-    mcmc_out$logLik <- mcmc_out$mlogLik <- NULL
+    if (!control$save_logLik_contributions) mcmc_out$logLik <- mcmc_out$mlogLik <- NULL
     c(mcmc_out, list(statistics = statistics,
                      fit_stats = list(conditional = conditional_fit_stats,
                                       marginal = marginal_fit_stats)),
